﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using Reloaded.Memory.Streams;

namespace FpkTool
{
    class Program
    {
        struct FpkHeader
        {
            public uint fpk;
            public uint zero1;
            public uint zero2;
            public uint count;
        }

        struct FileEntry
        {
            public int Hash;
            public int Length;
            public int Offset;
            public int Padding;
        }
        static void Main(string[] args)
        {
            // 循环直到用户退出
            while (true)
            {
                Console.WriteLine("请输入命令或路径（输入 'exit' 退出）：");
                string input = Console.ReadLine();

                if (input.ToLower() == "exit")
                {
                    // 用户输入 exit 时退出程序
                    Console.WriteLine("程序已退出.");
                    break;
                }

                // 根据用户输入处理
                var arguments = input.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);

                if (arguments.Length == 0)
                {
                    PrintUsage();
                }
                else if (arguments.Length == 2)
                {
                    CompareLua(arguments[0], arguments[1]);
                }
                else
                {
                    var path = Path.GetFullPath(arguments[0]);
                    if (File.Exists(path))
                    {
                        DumpFpk(path);
                    }
                    else if (Directory.Exists(path))
                    {
                        PackFpk(path);
                    }
                    else
                    {
                        PrintUsage();
                    }
                }
            }
        }

        static void PrintUsage()
        {
            // 输出程序名称
            Console.WriteLine("Fpk工具");
            // 输出使用说明：Fpk文件路径
            Console.WriteLine("用法: FpkTool <fpk文件路径>");
            // 输出使用说明：解压后的fpk文件夹路径
            Console.WriteLine("       FpkTool <提取后的fpk文件夹路径>");
            // 输出使用说明：未命名的fpk文件夹路径和命名的fpk文件夹路径
            Console.WriteLine("       FpkTool <未命名的fpk文件夹路径> <命名的fpk文件夹路径>");
            // 示例：运行程序的示例路径
            Console.WriteLine("例如: FpkTool.exe " + @"H:\pso2_bin\data\win32\00800bcb4e790060f5a47b85fbc2acd0");
        }

        static void DumpFpk(string filePath)
        {
            var fileDict = new Dictionary<int, string>();
            var storeDict = new Dictionary<int, byte[]>();
            var luaFiles = new List<string>();
            var refText = Path.Combine(Path.GetDirectoryName(Assembly.GetExecutingAssembly().GetName().CodeBase).Substring(6), "FpkTool.txt");

            //Get name references if they exist
            if (File.Exists(refText))
            {
                Console.WriteLine("FpkTool.txt 已存在");
                using (StreamReader reader = File.OpenText(refText))
                {
                    string line;
                    while ((line = reader.ReadLine()) != null)
                    {
                        string[] fileKey = line.Split(' ');
                        fileDict.Add(int.Parse(fileKey[0], System.Globalization.NumberStyles.HexNumber), fileKey[1]);
                    }

                }
            } else
            {
                Console.WriteLine("FpkTool.txt 不存在");
            }

            //Extract files and apply names if applicable
            Console.WriteLine("处理中...");
            if (File.Exists(filePath))
            {
                var fpk = File.ReadAllBytes(filePath);

                using (var fileStream = new FileStream(filePath, FileMode.Open))
                using (var streamReader = new BufferedStreamReader(fileStream, 8192))
                {
                    List<FileEntry> table = new List<FileEntry>();

                    // 跳过文件头的偏移量（0xC）并读取条目数量
                    streamReader.Seek(0xC, SeekOrigin.Begin);
                    streamReader.Read<UInt32>(out UInt32 entryCount);

                    // 读取文件条目
                    for (int i = 0; i < entryCount; i++)
                    {
                        streamReader.Read(out FileEntry value);
                        table.Add(value);
                    }

                    // 获取条目表的结束位置
                    long tableEnd = streamReader.Position();

                    // 遍历每个条目并读取对应的数据
                    foreach (FileEntry entry in table)
                    {
                        // 计算数据读取位置
                        long dataPosition = entry.Offset + tableEnd;
                        int dataLength = (int)entry.Length;

                        // 检查数据读取位置和长度是否有效
                        if (dataPosition >= 0 && dataPosition + dataLength <= fileStream.Length)
                        {
                            storeDict.Add(entry.Hash, streamReader.ReadBytes(dataPosition, dataLength));
                        }
                        else
                        {
                            Console.WriteLine($"警告：无效的读取位置或长度，Hash: {entry.Hash}");
                        }
                    }
                }

                var outputDirectory = Path.Combine(Path.GetDirectoryName(filePath), Path.GetFileNameWithoutExtension(filePath));
                Directory.CreateDirectory(outputDirectory + "_");
                using (StreamWriter fpkList = new StreamWriter(outputDirectory + ".txt"))
                {
                    foreach (var file in storeDict)
                    {
                        string name;
                        if (fileDict.ContainsKey(file.Key))
                        {
                            name = fileDict[file.Key];
                        } else
                        {
                            name = file.Key.ToString("X") + ".lua";
                        }
                        var fileOutput = Path.Combine(outputDirectory + "_", $"{name}");
                        File.WriteAllBytes(fileOutput, file.Value);
                        fpkList.WriteLine(file.Key.ToString("X") + " " + name);
                    }
                }
            }
        }

        static void PackFpk(string dirPath)
        {
            var refText = dirPath.Substring(0, dirPath.Count()-1) + ".txt";
            var fileDict = new Dictionary<int, byte[]>();

            //Get name references and file data; error if the file isn't there
            if (File.Exists(refText))
            {
                Console.WriteLine(Path.GetFileName(refText) + " exists");
                using (StreamReader reader = File.OpenText(refText))
                {
                    string line;
                    while ((line = reader.ReadLine()) != null)
                    {
                        string[] fileKey = line.Split(' ');
                        if(File.Exists(Path.Combine(dirPath, fileKey[1])))
                        {
                            fileDict.Add(int.Parse(fileKey[0], System.Globalization.NumberStyles.HexNumber), File.ReadAllBytes(Path.Combine(dirPath, fileKey[1])));
                        } else
                        {
                            Console.WriteLine("错误: 文件 " + fileKey[1] + " 不存在目录中!");
                            Console.ReadKey();
                            return;
                        }
                    }

                }
            } else
            {
                Console.WriteLine("错误: 目录词典 " + Path.GetFileName(refText) + " 不存在!");
                Console.ReadKey();
                return;
            }
            Console.WriteLine("处理中...");
            //Write fpk
            var fpkMem = new MemoryStream();

            //Write header
            var fileHeader = new FpkHeader();
            fileHeader.fpk = 7041126; fileHeader.zero1 = 0; fileHeader.zero2 = 0; fileHeader.count = (uint)fileDict.Count;

            var fileHeaderBytes = Reloaded.Memory.Struct.GetBytes(ref fileHeader, true);
            fpkMem.Write(fileHeaderBytes, 0, fileHeaderBytes.Count());

            var offset = 0;
            foreach (var f in fileDict)
            {
                var entry = new FileEntry();
                entry.Hash = f.Key;
                entry.Length = f.Value.Count(); 
                entry.Offset = offset;
                entry.Padding = 0;
                offset += f.Value.Count();

                //Pad to multiple of 4
                var offPadding = 4 - (offset % 4);
                if (offPadding < 4)
                {
                    offset += offPadding;
                }

                var entryBytes = Reloaded.Memory.Struct.GetBytes(ref entry, true);
                fpkMem.Write(entryBytes, 0, entryBytes.Count());
            }

            //Write luas
            int counter = 0;
            foreach (var f in fileDict)
            {
                fpkMem.Write(f.Value, 0, f.Value.Count());
                //Pad to a multiple of 4
                var padding = 4 - (fpkMem.Position % 4);
                if (padding < 4 && counter < fileDict.Count()-1)
                {
                    for (int i = 0; i < padding; i++)
                    {
                        fpkMem.WriteByte(0);
                    }
                }
                counter++;
            }

            //Write to file
            var newFile = dirPath + "new.ice";
            File.WriteAllBytes(newFile, fpkMem.ToArray());
        }

        static void CompareLua(string noNamePath, string namePath)
        {
            var filesNoName = Directory.GetFiles(noNamePath);
            var filesName   = Directory.GetFiles(namePath);
            var NoDict = new Dictionary<uint, string>();
            var NameDict = new Dictionary<uint, string>();

            //Gather names and hashes
            foreach (var f in filesNoName)
            {
                var fTest = Extensions.Data.XXHash.XXH32(File.ReadAllBytes(f), 0);
                if (NoDict.ContainsKey(fTest))
                {
                    Console.WriteLine("发现重复项");
                    Console.WriteLine(f);
                    Console.WriteLine(NoDict[fTest]);
                    NoDict[fTest] = NoDict[fTest] + " " + Path.GetFileName(f); //Record duplicate names so we can have some idea of what they might be
                } else
                {
                    NoDict.Add(fTest, Path.GetFileName(f));
                }
                
            }
            foreach (var f in filesName)
            {
                var fTest = Extensions.Data.XXHash.XXH32(File.ReadAllBytes(f), 0);
                if (NameDict.ContainsKey(fTest))
                {
                    Console.WriteLine("发现重复项");
                    Console.WriteLine(f);
                    Console.WriteLine(NameDict[fTest]);
                    NameDict[fTest] = Path.GetFileNameWithoutExtension(NameDict[fTest]) + "_OR_" + Path.GetFileName(f); //Record duplicate names so we can have some idea of what they might be
                }
                else
                {
                    NameDict.Add(fTest, Path.GetFileName(f));
                }
            }

            //Compare hashes and write back with applied names if applicable
            var refText = Path.Combine(Path.GetDirectoryName(Assembly.GetExecutingAssembly().GetName().CodeBase).Substring(6), "FpkTool_Named.txt");
            using (StreamWriter fpkList = new StreamWriter(refText))
            {
                foreach (var f in NoDict)
                {
                    var name = NoDict[f.Key];
                    if (NameDict.ContainsKey(f.Key))
                    {
                        name = NameDict[f.Key];
                    }
                    var fileList = NoDict[f.Key].Split(' ');
                    if (fileList.Count() > 1)
                    {
                        for (int i = 0; i < fileList.Count(); i++)
                        {
                            fpkList.WriteLine(Path.GetFileNameWithoutExtension(fileList[i]) + " " + i.ToString() + name);
                        }
                    } else
                    {
                        fpkList.WriteLine(Path.GetFileNameWithoutExtension(fileList[0]) + " " + name);
                    }
                }
            }
        }
    }
}
